/*******************************************************************
 *                                                                 *
 *		        	C format converter        CFV.C						    *
 *                                                                 *
 *               Copyright 1990 Ron Mignery                        *
 *                                                                 *
 *******************************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#define TRUE 1
#define FALSE 0

FILE *fpi, *fpo;
char line_buff[64][128], outline[256], infile[64], tempfile[64],
 outfile[64], tabline[80], in_char;
int indent_in_debug, in_debug, new_line_before_lbrace, lineno = 0,
 use_no_tabs, exempt, next_exempt, no_space_indent, echo_to_screen;
int new_line, st_end, st_cont, exempt, idx, fungetc, previous_char;

void main (int argc, char *argv[]) {

	int i, j, k, l, incom, paren, squotes, quotes, opts, add, overwrite;
	int tab[40], outdent_after_rbrace, indent_before_lbrace,
	 no_indent_col_1, line_end, count, count_adjust;

	if (argc < 2) {              /* if no command line arguements */
		printf("\nCFV - C fixed format converter v1.0\n");
		printf
		("\nFormat is:  cfv [-eilmnost] filename.c [outfile.c] [tab] [tab...]\n\n");
		printf("\te = format in DEBUG lines\n");
		printf("\ti = indent before lbrace ({)\n");
		printf("\tl = newline before lbrace\n");
		printf("\tm = no space indent for multi-line statements\n");
		printf("\tn = no indent on column 1\n");
		printf("\to = outdent after rbrace (})\n");
		printf("\ts = echo output to screen\n");
		printf("\tt = use spaces instead of tabs\n");
		exit(0);
	}

	indent_in_debug = new_line_before_lbrace = use_no_tabs = echo_to_screen
	 = outdent_after_rbrace = indent_before_lbrace = no_indent_col_1 = FALSE;
	opts = 0;
	while (argv[opts+1][0] == '-') {    /* scan options */
		opts++;
		if(strchr(argv[opts],'e')) indent_in_debug = TRUE;
		if(strchr(argv[opts],'i'))
		 indent_before_lbrace = new_line_before_lbrace = TRUE;
		if(strchr(argv[opts],'l')) new_line_before_lbrace = TRUE;
		if(strchr(argv[opts],'m')) no_space_indent = TRUE;
		if(strchr(argv[opts],'n')) no_indent_col_1 = TRUE;
		if(strchr(argv[opts],'o')) outdent_after_rbrace = TRUE;
		if(strchr(argv[opts],'s')) echo_to_screen = TRUE;
		if(strchr(argv[opts],'t')) use_no_tabs = TRUE;
	}

	printf("\n");

	strcpy(infile, argv[1+opts]);                /* scan input filename */
	if ((fpi = fopen(infile, "r")) == NULL) {
		printf("\nFile not found\n");
		exit(-1);
	}

	if ((argc > (2+opts)) && (atoi(argv[2+opts])== 0)
	 && (strcmp(infile, argv[2+opts]) != 0)) {   /* scan output filename */
      opts++;
      overwrite = FALSE;
		strcpy(outfile, argv[1+opts]);
   }
   else {
		overwrite = TRUE;     /* overwrite input file if no output filename */
   	strncpy(outfile, infile, (strchr(infile,'.')-infile));
      strcat (outfile,".tmp");
   }

	if ((fpo = fopen(outfile,"w")) == NULL) {
		printf("\nUnable to create output file\n");
		exit(-2);
	}

	printf("C Format Conversion on %s\n\n", infile);

	tab[0] = 1;                         /* set tabs */
   for (i=1; i< argc-1-opts; i++) {
		tab[i] = atoi(argv[i+1+opts]);
      if (tab[i] <= tab[i-1]) {
		   printf("Tab sequence error\n");
		   exit(-3);
	   }
   }
   if (i>1) add = tab[i-1]-tab[i-2];
   else add = 3;

   for (;i < 40; i++) tab[i] = tab[i-1] + add;

	if (use_no_tabs) strcpy(tabline,"                                                  ");
	else strcpy(tabline,"\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t");

	quotes = squotes = in_debug = incom
	 = line_end = new_line = st_end = st_cont = exempt = FALSE;
	idx = paren = count = count_adjust = fungetc = i = 0;
	new_line = TRUE;

	while (feof(fpi) == FALSE) {        /* scan input file */

		if (previous_char == '\\') previous_char = 0;
		else previous_char = in_char;

		if (fungetc) {                   /* simulate unget for file */
			in_char = fungetc;
			fungetc = 0;
		}
		else in_char = fgetc(fpi);       /* get char from file */

		/* if end of statement reached, analyze here */
		if (st_end) switch(in_char) {
		case ' ':                        /* LF in whitespace if */
		case '\t':                       /*  line is too long */
		case '\n':
         if (idx > 76) in_char = '\n';
			break;
		case '/':                        /* scan for end of comment */
			if (idx && (outline[idx-1] == '*')) incom = FALSE;
			break;
		case '*':                        /* scan for start of comment */
			if (idx && (outline[idx-1] == '/')) incom = TRUE;
			break;
      default:
			if (incom == FALSE) {         /* if not in comment or whitespace */
				fungetc = in_char;         /* insert LF in char stream */
				in_char = '\n';
			}
		}

		/* else if in comment, analyze here */
		else if (incom) {                /* scan for end of comment */
			if ((in_char == '/') && (previous_char == '*')) incom = FALSE;
			if (idx > 80) line_end = TRUE;/* LF if line too long */
		}

		/* else if in quotes, analyze here */
		else if (quotes) {               /* scan for end quote */
			if ((in_char == '\"') && (previous_char != '\\')) quotes = FALSE;
			if (idx > 80) {               /* insert backslash LF if line too */
				fungetc = in_char;         /* long */
            in_char = '\\';
            line_end = TRUE;
         }
		}

		/* else if in single quotes, analyze here */
		else if (squotes) {              /* scan for end quote */
			if ((in_char == '\'') && (previous_char != '\\')) squotes = FALSE;
		}

		/* else, analyze here */
		else switch(in_char) {
		case ' ':                        /* LF in whitespace if */
		case '\t':                       /*  line is too long */
         if (idx > 76) in_char = '\n';
			break;
		case ';':                        /* ; => end of statement */
			if (paren == 0) st_end = TRUE;/* unless in parentheses */
			break;
		case '\"':                       /* start of literal unless \" */
			if ((in_char == '\"') && (previous_char != '\\'))
			quotes = TRUE;
			break;
		case '\'':                       /* start of literal unless \' */
			if ((in_char == '\'') && (previous_char != '\\'))
			squotes = TRUE;
			break;
		case '{':                        /* { => end of statement */
			st_end = TRUE;                /* start of block */

													/* if preceeding text */
			if (strspn(outline," \t") < idx) {
				if (new_line_before_lbrace) {
					fungetc = in_char;      /*  indent according to options */
					in_char = '\n';
				}
				else count_adjust++;
			}

			else {                        /* if no preceeding text */
				st_cont = FALSE;           /*  disable space indent */
				if (indent_before_lbrace) count++;
				else count_adjust++;       /*  indent according to options */
			}

			break;
		case '}':                        /* } => end of block */
			st_end = TRUE;                /* outdent according to options */
			if (outdent_after_rbrace) count_adjust--;
			else count--;
			break;
		case '#':                        /* no indent for directives */
			if (new_line) exempt = TRUE;
			break;
		case '*':                        /* scan for start of comment */
			if (idx && (outline[idx-1] == '/')) incom = TRUE;
			break;
		case '(':                        /* track level of parenthesis */
			paren++;
			break;
		case ')':                        /* track level of parenthesis */
			paren--;
			break;
		}

		if (in_char == '\n') line_end = TRUE;
		else outline[idx++] = in_char;   /* add char to string */

		if (line_end) {                  /* if end of a line */


			/* don't indent next line if this one ends with \ */
			if (idx && outline[idx-1] == '\\') next_exempt = TRUE;

			/* don't space indent after labels */
			if (strchr(outline,':') != NULL) st_end = TRUE;

			while (idx &&                 /* remove trailing white space */
			 ((outline[idx-1] == ' ') || (outline[idx-1] == '\t'))) idx--;
			outline[idx] = '\0';

			/* detect start of debug lines */
			if (strncmp(outline,"#ifdef DEBUG", 12) == 0) in_debug = TRUE;

			/* don't indent debug lines if option not set */
			if (in_debug && (indent_in_debug==FALSE)) exempt = TRUE;

			/* detect end of debug lines */
			if (in_debug && (strncmp(outline,"#endif", 6) == 0)) in_debug = FALSE;

			j=0;                          /* find first non-white space */
			while ((outline[j] == ' ') || (outline[j] == '\t')) j++;

			/* don't indent full line comments or blank lines */
			if (strncmp(&outline[j],"/*",2) == 0) exempt = TRUE;
			if (strlen(outline) < 1) exempt = TRUE;

			if (exempt) {                 /* if indenting disabled */
				if (lineno > 63) {
					fputs(line_buff[lineno % 64], fpo);
					if (echo_to_screen) printf("%s", line_buff[lineno % 64]);
				}
				sprintf(line_buff[lineno % 64],"%s\n",outline);
			}

			else {                        /* else */

				/* remove leading whitespace */
				if (j) strcpy(outline,&outline[j]);

				/* if statement continuation and option not set */
				if ((no_space_indent == FALSE) && (st_cont)) {
					movmem(outline, &outline[1], strlen(outline)+1);
					outline[0] = ' ';       /*  space indent */
            }
				st_cont = st_end ? FALSE : TRUE; /* update continuation flag */

				/* compute indentation */
				k = count - no_indent_col_1;
				if (k < 0) k = 0;
				if (!use_no_tabs) j = k;
				else j = tab[k]-1;

				/* output line */
				if (lineno > 63) {
					fputs(line_buff[lineno % 64], fpo);
					if (echo_to_screen) printf("%s", line_buff[lineno % 64]);
				}
				sprintf(line_buff[lineno % 64], "%.*s%s\n", j, tabline, outline);
			}

         /* pull up lbrace to right of code unless new line before lbrace */
		   if ((new_line_before_lbrace == FALSE)
          && (line_buff[lineno % 64][strspn(line_buff[lineno % 64]," \t")]
          == '{')) {

            for (i = lineno-1; i>= 0; i--) {
               for (j = l = strlen(line_buff[i%64])-1; j>=0 ; j--) {
                  switch (line_buff[i%64][j]) {
                  case '/':
                     if (j && (line_buff[i%64][j-1] == '*')) incom = TRUE;
                     if (line_buff[i%64][j+1] == '*') incom = FALSE;
                  case ' ':
					   case '\t':
					   case '\n':
                     break;
                  default:
                     if (incom == FALSE) {
							   movmem(&line_buff[i%64][j+1], &line_buff[i%64][j+3],
							   l-j+1);
                        line_buff[i%64][j+1] = ' ';
                        line_buff[i%64][j+2] = '{';

                        /* keep comments */
								if (strchr(line_buff[lineno % 64], '/') != NULL) {
									line_buff[i%64][l+2] = '\0';
                           strcat(line_buff[i%64],
									 1+strchr(line_buff[lineno % 64], '{'));
                        }

		                  line_buff[lineno % 64][0] = '\0';
                        j = i = -1;
                     }
                  }
               }
            }
         }

		   lineno++;

			if (line_end) st_end = 0;     /* update flags */
			line_end = idx = 0;
			new_line = TRUE;
			exempt = incom | next_exempt;
			next_exempt = FALSE;
			count += count_adjust;
			count_adjust = 0;
		}

		else new_line = FALSE;

	}

	if (lineno < 64) for (i=0; i<lineno; i++) {
		fputs(line_buff[i], fpo);
		if (echo_to_screen) printf("%s", line_buff[i]);
	}
	else for (i=0; i<64; i++) {
		fputs(line_buff[(lineno+i) % 64], fpo);
		if (echo_to_screen) printf("%s", line_buff[(lineno+i) % 64]);
	}

	fcloseall();                        /* close all files */

	if (overwrite) {                    /* cleanup if overwriting input file */
   	strncpy(tempfile, infile, (strchr(infile,'.')-infile));
		strcat(tempfile,".bak");
		remove(tempfile);
		rename(infile, tempfile);
		rename(outfile, infile);
   }
}

